/*********************************************************
 * DiodeCalibration skeleton    whmp 2015 11 16 
 *
 * When compiled for a 168 based Arduino the Marquee and indicator font are removed.
 *
 * Calibration of a diodes bandgap voltage in respect to temperature and VCC
 * (for use as accurate temperature sensors). The relationship is highly linear,
 * the calibrated temperature is a simple mx+b , m and b being the calibration
 * results and x an adc diode measurement. Once calibrated do not change diode 
 * resistor pairing or supply voltage. The datums for temperature, vcc, and the diode
 * readings are output on the serial port for external logging. For best results good 
 * 'lab practices' as well as electronics are required.
 * 
 * control keys
 * 
 * t thermister master reference - default adc 0
 * d diode master reference - adc 0 (previously calibrated)
 * u User master reference - user enters temperature manually
 * r reset
 * c auto calibrate
 * p pause
 * m log manual sample point
 * <enter> set manual data
 * <backspace> clear manual data
 * 
 * As a MRETV learning experience ...
 * 
 * The CAPSfont is a macro.
 * #define CAPSfont ((prog_char* )((uint)CAPS_font-(32*8)) )
 * Its a false pointer to use a tiny 56 char font (points to 32 characters before the font)
 * 
 * The screens scrolling Marquee is dual purposed as a buffer to format the serial output data. Look
 * closely and you can see it happen during the beep. A lot of screen data is dual purposed as
 * the record of samples taken while displaying that information to the user.
 *
 * The background ADC routine calculates a real average,
 * not an average x ADC_SAMPLECOUNT as was done in keyboard tool.
 *
 * One sample per 1/10 degree celsius over 120 degree range = 1200 sample points for each diode.
 * 50 ADC reads for each sample, Accumulated data 50*1200 = 60,000 ADC measurement for each diode
 * (maximum). With 4 diodes and the temperature reference and VCC data thats 242,400 (~1/4 million) ADC 
 * measurements, and it beeps. In manual mode you can get usable results from 3+ or more data points if you  
 * keep your data points evenly spread over the intended use range. 
 *
 * There is a GOPL that shows some interesting things. No need to try to understand it completely.
 *
 * About Extension defines.
 *
 * Defining these in the MRETV library folder allows for extra optimisation and
 * improved performance. All examples are predefined in Arduino\libraries\MRETV\MRETVextensions.h
 *
 * for MasteRasterExtendedTransmissionVision 2
 **********************************************************/

#include <MrETV.h>
#ifndef DiodeCalibration // make sure the extensions are set correctly
//#error Please define DiodeCalibration in arduino\libraries\MRETV\MRETVextensions.h , See comments at start of sketch
#endif
/*****************
 * #define firmwaremultiport
 * #define translatekeys
 * #define mergstreams
 * #define userextension   background()
 * #define fontheight 8
 * #define graphicsengine
 * #define countframes
 * #define monoaudio
 * #define ADC_SAMPLECOUNT 50
 ********************/

#include "CAPSfont.h"
#include "saw.h"
#include "linefont.h"  // fancy font and sound for 328 version

// #define __AVR_ATmega168__   // for testing only

#ifdef __AVR_ATmega168__
#define linefont CAPSfont  // hide the indicator font.
#endif

// for a thermister master reference (set for your own)
#define SH_A .003354016
#define SH_B .0002569850
#define SH_C  .000002620131
#define SH_D  .00000006383091

// real value of the thermister's resistor  (measure don't assume)
#define hwohms 9940

// for A Diode master reference (set for your diode/resistor/vcc combination)
#define refM -.4538L
#define refB 5277L

// measured ratio of  VCC / voltage divider sensor
#define hwcompvolt  (6.906)

uchar  Screenbuf[416-18]; // highly GOPed to save memory, swap fonts ,bold data, scroll SINGLE text line ..
volatile uint analog[12]; // adc values and averages
uint analogindex;  // counter

// user interupt routine
void background(){  // read adc's and take and store an average of ADC_SAMPLECOUNT samples
  uchar a;
  if(( Video.LineCount> Video.ScreenBottom)) { // once a frame
    union{  // for reading adc
      short val;
      uint8_t val8[2];
    };
    if(ADCSRA&_BV(ADSC)) return;// too early better miss a sample
    val8[0]=ADCL; // read conversion results
    val8[1]=ADCH;

    a=analogindex%8; //repeat 8 steps
    switch(a){
    case 0: // first datum after ref change, ignore.
      break;
    case 1:
      analog[0]+=val; // store 0
      ADMUX=ADC_11V|1; // next  1 1v (change of vref)
      break;
    case 2: // first after ref change ignore.
      break;
    case 3: //store 1 next 2 1v
    case 4: //store 2 next 3 1v
    case 5: //store 3 next 4 1v
    case 6: //store 4 next 5 1v
      analog[a-2]+=val;
      ADMUX=ADC_11V|(a-1);
      break;
    case 7:
      analog[5]+=val; //store 5
      ADMUX=ADC_5V|0; //next 0 5v (change of vref)
    }
    ADCSRA=ADC_START; // start next conversion
    analogindex=(analogindex+1)%(8*ADC_SAMPLECOUNT); // increment count
    if((analogindex<8)&&(analogindex>1) ){ // update totals
      // analog[analogindex+4]=analog[analogindex-2]; // this would give average * ADC_SAMPLECOUNT
      analog[analogindex+4]=(analog[analogindex-2]+(ADC_SAMPLECOUNT/2)-1)/(ADC_SAMPLECOUNT);
      analog[analogindex-2]=0; // reset for new total
    }
  }
}

// This is an over the top GOPL example. You can see lines of 1 or 2 characters in length
// and special fonts for certain lines. These are useful ideas. The bit scroll stuff is
//only for those with the purest of hearts.


#define caltop 20
volatile uchar calibrategopl[]={
  caltop+7,flag_gopcharwidth|flag_gopscreenwidth|flag_gopcharheight|flag_gopcenter,1,10,2,44,
  caltop+23,flag_gopcharwidth|flag_gopscreenwidth|flag_gopcharheight|flag_gopcenter,   0,   12,  1, 75,
  caltop+31,flag_gopscreenwidth,25,
  caltop+39,flag_gopscreenwidth,5,
  caltop+47,flag_gopscreenwidth,25,
  caltop+55,flag_gopscreenwidth,20,
  caltop+63,flag_gopscreenwidth,25,
  caltop+95,flag_gopscreenwidth,7,
  caltop+95+8,flag_gopscreenwidth,3,
  caltop+103+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+111+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),1,
  caltop+119+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+127+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),2,
  caltop+135+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+143+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),2,
  caltop+151+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+159+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),2,
  caltop+167+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+175+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),2,
  caltop+183+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(linefont),25,
  caltop+191+8,flag_gopcharset|flag_gopscreenwidth ,gop_int(CAPSfont),6,
  255,flag_gopscreenwidth|flag_gopcenter|flag_gopreset,25, /**/96/**/ ,gop_int(calibrategopl)
  };

#ifdef __AVR_ATmega168__  // no Marquee
#else
#define helplenth 128
const unsigned char calibratehelp[helplenth+1] PROGMEM = {
  "MASTERRASTEREXTENDEDTRANSMISSIONVISION REFERENCE T THERMISTOR D DIODE U USER  COMMAND C CALIBRATE/CONT P PAUSE M MANUAL R RESET "};

// The cliffs of insanity.
// To bit scroll a line use the functions to GOP on the raster line preceding the textline you want
// GOP on the final raster line of the textline you  scrolled back to standard centering
// to scroll line 1 (raster 8-15) gop on line 7 and 15. GOPS are executed after rendering  in a  rasterline.
// In this example  since its text line 0 getting scrolled Gops are at screenbottom  and 7.
// These lines also how many characters are rendered flag_gopscreenwidth. This can save ALOT of ram.
// but means some extra effort in using print functions. Raster lines 7 and 23 are gopped for char size.
// the font is swapped back and forth to give pixel level access for the temperature samples map
// gop_int() is just a macro to make a  low,high from an address.
// btw ...the CAPSfont is another macro its a false pointer to use a tiny 56 char font
void scrollliner(){
  if((Video.FrameCount&7)==1){
    Screen.scrolllineright(0);
    Screen.write(0,0,pgm_read_byte(calibratehelp+(helplenth-1-((Video.FrameCount%(helplenth*8))/8))));
  }
  calibrategopl[96]=72+((Video.FrameCount%8)); // slide bits
}

void scrolllinel(){
  if((Video.FrameCount&7)==1){
    Screen.scrolllineleft(0);
    Screen.write(0,Screen.width-1,pgm_read_byte(calibratehelp+((Video.FrameCount%(helplenth*8))/8)));
  }
  calibrategopl[96]=72+(7-(Video.FrameCount%8)); // slide bits
}
#endif
// This is a project under development but I it is in usable shape and the results acurate.

struct student{
  double XY,Y;
};
struct student students[4];

//double X,Y,XY,X2,N,slope,Y3,Y4,Y5,XY3,XY4,XY5,usertemp;
double N,slope,B,usertemp,X,X2;

// the master temperature reference type
uchar reftype='T';


// temperature from diode sensor
double diode(double TEMPM,double TEMPB ,uchar adcsel){
  return((analog[6+adcsel]-TEMPB)/TEMPM);
}

// temperature from diode sensor
double newdiode(double TEMPM,double TEMPB ,uchar adcsel){
  return((analog[6+adcsel]*(TEMPM))+TEMPB);
}

// temperature from thermister sensor
double thermister(uint adcin){
  double temp=adcin;
  temp= hwohms / (1023/temp-1) ; // adc->ohms
  temp=log(temp/10000); //log(R/Rref)
  temp=(1/(SH_A+(SH_B*temp)+(SH_C*temp*temp)+(SH_D*temp*temp*temp)));//extended Steinhart and Hart
  return(temp); // -273.15 in celcius

}
//read voltage sensor
double volts(uint adcin,uint refvolts){
  double temp=adcin;
  temp=(temp/1023)*refvolts;
  return(temp);
}
double regression(uchar stud){
  slope=(N*X2-(X*X))/(N*students[stud].XY-(X*students[stud].Y));
  B=((students[stud].Y- (X/slope))/N)*(-slope) ;
  return(slope);
}

// add a data point to the regression series of each new diode
void adddatum(double datum){
  uchar a;
  if((samplebar(datum)))return;
  N=N+1;
  X+=datum;
  X2+=datum*datum;
  for(a=0;a<4;a++){
    students[a].XY+=datum*analog[2+6+a];
    students[a].Y+=analog[2+6+a];
  }
  Video.stringS(Screen.dec( 0,10,  datum    ));
  Video.writeS(',');
  Video.stringS(Screen.udec(0,10, volts(analog[1+6],1100)*hwcompvolt     ));
  Video.writeS(',');
  for(a=0;a<4;a++){
    Video.stringS(Screen.dec( 0,10,  analog[2+6+a]     ));
    Video.writeS(',');
  }
  Video.stringS("\r\n");
  Screen.dec(0,10,thermister(analog[2+6])*10);
  update();
}


void setup(){
}

void loop(){
  uint input,a;
  uchar chin,learn=0;
  double temp=-1,lasttemp=-1,tusertemp=0;
  uchar frame,tt=0,incount=0;
  frame=Video.FrameCount;
  for(a=0;a<4;a++){
    students[a].XY=0;
    students[a].Y=0;
  }
  N=0;
  Screen.begin(CAPSfont,Screenbuf,26*fontheight,25, caltop,72,1);
  setgoplist(calibrategopl);  // twist up the screenbuf data
  Screen.fill(12,92,' ',-6); // clear  input line
  Screen.fillrow(0,' '); // clear  input line
  input=Screen.curloc; // on screen input buffer , strange screen location , display is gopped
  update();
  samplebar(0); // clear buffer
  Video.stringS(",\r\n");
  while(1)  {
    while(Video.LineCount) ;
#ifdef __AVR_ATmega168__
#else
    scrolllinel();
#endif
    if(frame!=(uchar)(Video.FrameCount/4)){ //64
      frame=(Video.FrameCount/4);  //64
      temp=reftemp(reftype);
      update();
      if((learn) &&(lasttemp!=temp)){
        lasttemp=temp;
        adddatum(temp);
      }
    }
    if((chin=Video.readS())){
      if((chin>='0')&&(chin<='9')){
        Screen.curloc=input;
        Screen.write(chin);
        if (incount++<5)input=Screen.curloc;
        tusertemp*=10;
        tusertemp+=chin-'0';
      }
      else switch (Screen.lower(chin)){
      case '.':
        // if(chin=='.'){
        Screen.curloc=input;
        Screen.write(chin);
        if (incount++<5)    input=Screen.curloc;
        tt|=1;
        break;
      case '-':
        tt|=2;
        Screen.curloc=input;
        Screen.write(chin);
        if (incount++<5)      input=Screen.curloc;
        break;
      case 0x0d:
        if((!tt&1)||(*(Screen.poke-1)=='.'))tusertemp*=10;
        if(tt&2){
          tusertemp*=-1;
          usertemp=tusertemp+2731; // ...round down ok
        }
        else usertemp=tusertemp+2731.5;  //round up
       // break; go ahead and clear!
      case 8: // backspace to clear
        tt=0;
        tusertemp=0;
        Screen.fill(12,92,' ',-6); // clear  input line
        input=Screen.curloc; // on screen input buffer , reset input while we are at it.
        incount=0;
        break;
      case 'm':
        adddatum(temp);
        break;
      case 'p':
        ssample(sawCo5,10);
        learn=0;
        break;
      case 'c':
        ssample(sawCo5,10);
        learn=1;
        break;
      case 'r':
        asm volatile("JMP 0x0000\n\t"::);
      case 't':
        reftype='T';
        break;
      case 'd':
        reftype='D';
        break;
      case 'u':
        reftype='U';
        break;
      }
    }
  }
}

// lines of diff len 25,15,12,25,5,25,20,25,25,25,25,25,7,3,25,1,25,2,25,2,25,2,25

void update()  {  // screen is gopped to save memory ..the display is still simple. 'screen' thinks  25 char line
  uchar a;
  uint temp;
  // Screen_progstr(0,0, " REF = T D U  CMD = C P M R ");
  Screen.setcurloc(1,0); // 2nd line
  temp=reftemp(reftype);//thermister(analog[0+6])*10;
  Screen.fill(' ',-10); // 10 spaces length of this line, negative leaves pointer at same place ( start of 2nd line).
  Screen.dec(temp-2731.5 ); // 2nd line
  Screen.skip(2);
  Screen.udec(volts(analog[1+6],1100)*hwcompvolt);
  Screen_progstr(0,35, "RAW ADC DATA"); // force to postion 30
  while(Video.LineCount<100) ; //flicker fixer
  Screen.fill(' ',-25); // clears a line
  Screen.udec((analog[0+6]));
  Screen.skip(1);
  Screen.udec((analog[1+6]));
  Screen.skip(1);
  Screen.udec((analog[2+6]));
  Screen.skip(1);
  Screen.udec((analog[3+6]));
  Screen.skip(1);
  Screen.udec((analog[4+6]));
  Screen.skip(1);
  Screen.udec((analog[5+6]));
  Screen_progstr(0,72, "VOLTS");
  Screen.fill(' ',-25); // clears a line
  Screen.dec(volts(analog[2+6],11000));
  Screen.skip(2);
  Screen.dec(volts(analog[3+6],11000));
  Screen.skip(2);
  Screen.dec(volts(analog[4+6],11000));
  Screen.skip(2);
  Screen.dec(volts(analog[5+6],11000));
  Screen_progstr(0,102, "CALIBRATION N " );
  Screen.fill(' ',-6); // clears a line
  Screen.udec(N);
  for(a=0;a<4;a++){
    temp=regression(a);
    Screen.fill(4+a,22,' ',-25); // clears a line
    Screen.dec(4+a,22,slope*1000); // same loc to start output
    Screen.dec(4+a,30,B);
    Screen.dec(4+a,40,newdiode(slope,B,2+a)-2731.5 );
  }
  Screen_progstr(8,22, "SAMPLES");  // 8 lines =200, +17 =curloc 217
  Screen_progstr(8,29, "-20");
  Screen_progstr(8,57, "0");
  Screen_progstr(8,83, "20");
  Screen_progstr(12,10, "40");  //Remember these are uchar max 255, rollover cols to line every 100 to be safe. rollover =+4 lines
  Screen_progstr(12,37, "60");
  Screen_progstr(12,64, "80");
  // screen 416 chars,  chars gop is 96 chars (-26 for fonts/size work) =70,total 486 bytes for a 25*26=650 Screen. saved 165 bytes.
  //positive side effect: removes aprox 32 clock /char_reduction/ frame @  60 frames/sec (316800 cycles/sec more non render time).
  //...while i'm talking... actually 150 bytes are the bit array of the 1200 data points for calibration, it just happens to show
  // on screen but should not be considered used. (the sketch may change a bit with out these figures being updated)
  Screen.write(12,91,reftype);
}

uchar samplebar(int temp){
  uint base;
  uchar used;
  if(!temp){ // clear buffer... I assume we will not achieve absolute 0 until after calibration.
    for (used=0;used<25;used++){
      Screenbuf[used+232]=0;
      Screenbuf[used+258]=0;
      Screenbuf[used+285]=0;
      Screenbuf[used+312]=0;
      Screenbuf[used+339]=0;
      Screenbuf[used+366]=0;
    }
  }
  temp=temp-2731.5+200; // translate our range now 0-1199 is actually -20 to 100
  if((temp>1199)||(temp<0))return(1);  //don't use sample out of range
  switch(temp/200){ // find our screen row
  case 0:
    base=232;
    break;
  case 1:
    base=258;
    break;
  case 2:
    base=285;
    break;
  case 3:
    base=312;
    break;
  case 4:
    base=339;
    break;
 default:
 case 5:
    base=366;
    break;
  }
  temp=temp%200; // offset into row
  if(!(   used=Screenbuf[base+(temp/8)]&_BV(7-(temp&7))    )) // check datum
    Screenbuf[base+(temp/8)]|=_BV(7-(temp&7)) ;  //mark as used
  if(!used)ssample(sawCo6,10);
    return(used); // sample was already done
}

double reftemp(uchar mode){
  switch(mode){
  case 'T':
    return((thermister(analog[0+6])*10)+.5); // thermister on adc 0
  case 'D':
    return ((newdiode(refM,refB,0))+.5); //  previously calibrated diode
  default:
  case 'U':
    return(((double)usertemp)+.5); // Manually entered temperature
  }
}











